"""A module defining functions & objects that handle common computational geometry algorithms"""

import numpy as np


def is_ccw(a, b, c):
    """
    Determines if elbow formed by points ABC is counter clockwise (ccw).

    Parameters
    ----------
    a : tuple
        Point A as (x, y)
    b : tuple
        Point B as (x, y)
    c : tuple
        Point C as (x, y)

    Returns
    -------
    lgi : logical np.ndarray
        True if elbow ABC is counter clockwise
    """

    # is angle AC > angle AB ?
    # is slope AC > slope AB ?
    lhs = (c[1] - a[1]) * (b[0] - a[0])
    rhs = (b[1] - a[1]) * (c[0] - a[0])
    return lhs > rhs


def is_intersection(a, b, c, d):
    """
    Determines if polyline AB crosses polyline CD.

    Parameters
    ----------
    a : tuple
        Point A as (x, y)
    b : tuple
        Point B as (x, y)
    c : tuple
        Point C as (x, y)
    d : tuple
        Point D as (x, y)

    Returns
    -------
    lgi : logical np.ndarray
        True if line AB crosses line CD
    """

    abc = is_ccw(a, b, c)
    abd = is_ccw(a, b, d)

    cda = is_ccw(c, d, a)
    cdb = is_ccw(c, d, b)

    return np.logical_and(abc != abd, cda != cdb)


def get_intersection(a, b, c, d):
    """
    Gets parameters for intersections of line\s AB and line\s CD.

    Parameters
    ----------
    a : tuple
        Point A as tuple (x, y) of line\s AB
    b : tuple
        Point B as tuple (x, y) of line\s AB
    c : tuple
        Point C as tuple (x, y) of line\s CD
    d : tuple
        Point D as tuple (x, y) of line\s CD

    Returns
    -------
    x : tuple (t, u, lc, rc, lgi)
        Line intersection parameters
    """

    # Basic preparation
    x2_x1 = b[0] - a[0]
    y2_y1 = b[1] - a[1]
    x4_x3 = d[0] - c[0]
    y4_y3 = d[1] - c[1]
    x1_x3 = a[0] - c[0]
    y1_y3 = a[1] - c[1]

    # Determine numerator a, numerator b & denominator
    num_a = x4_x3 * y1_y3 - y4_y3 * x1_x3
    num_b = x2_x1 * y1_y3 - y2_y1 * x1_x3
    denom = (y4_y3 * x2_x1 - x4_x3 * y2_y1)

    # Check for parallel lines
    is_parallel = denom == 0
    denom[is_parallel] = -999

    # Get polyline intersection parameters
    t = np.round(num_a * (1 / denom), 8)  # Normalized distance along polyline AB
    u = np.round(num_b * (1 / denom), 8)  # Normalized distance along polyline CD
    lc = (x2_x1 * y4_y3) < (x4_x3 * y2_y1)  # Left cross if polyline ABC is counter clockwise
    rc = (x2_x1 * y4_y3) > (x4_x3 * y2_y1)  # Right cross if polyline ABD is counter clockwise

    # Set parallel lines to inf
    t[is_parallel] = np.inf
    u[is_parallel] = np.inf

    # Determine logical index for valid polyline intersection
    lgi = (0 <= t) & (t <= 1) & (0 <= u) & (u <= 1)

    return t, u, lc, rc, lgi


def in_polygon(x, y, xp, yp):

    a, b = (x, y), (x-10**8, y-10**8)

    winding_num = np.zeros(x.size, dtype=np.int32)
    for ii in range(xp.size-1):
        c, d = (xp[ii], yp[ii]), (xp[ii+1], yp[ii+1])
        lgi = is_intersection(a, b, c, d)

        if np.any(lgi):
            a_ii = (a[0][lgi], a[1][lgi])
            b_ii = (b[0][lgi], b[1][lgi])
            rc = 1*is_ccw(a_ii, b_ii, d)
            lc = -1*is_ccw(a_ii, b_ii, c)
            winding_num[lgi] = winding_num[lgi] + lc + rc
        else:
            continue

    return winding_num != 0


def get_unit_vectors(xv, yv):
    """
    Gets unit tangent and unit normal vectors for line segments

    Parameters
    ----------
    xv : 1D np.ndarray
        X coordinate of polyline
    yv : 1D np.ndarray
        Y coordinate of polyline

    Returns
    -------
    uv : tuple (ut, un)
        Unit tangent and unit normal vectors
    """

    dx = np.diff(xv)
    dy = np.diff(yv)
    ds = np.hypot(dx, dy)

    ut = np.transpose(np.vstack((dx/ds, dy/ds)))
    un = np.hstack((-ut[:, [1]], ut[:, [0]]))

    return ut, un


class Mesh:
    """
    Creates object which handles common 2D polygonal mesh based geometry operations

    Parameters
    ----------
    node_x : 1D np.ndarray
        X coordinate of mesh vertices
    node_y : 1D np.ndarray
        Y coordinate of mesh vertices
    cell_node : 2D np.ndarray
        A (n, 4) array defining each mesh cell/element by four node indices

    Attributes
    ----------
    nc2 : int
        Number of 2D mesh cells
    nv2 : int
        Number of 2D mesh vertices
    is_tri : 1D np.ndarray
        Logical index of triangular elements
    is_quad : 1D np.ndarray
        Logical index of quadrilateral elements
    edge_node : tuple
        Tuple defining start node, end node and cell for each mesh half edge
    weights : 2D np.ndarray
        A (n, 4) array defining weighting of each cell gives to each mesh vertex
    tri_cell_node : 2D np.ndarray
        A (n, 3) array defining each mesh cell/element by three node indices
    tri_cell_index : 1D np.ndarray
        A (n,) array mapping triangular mesh elements to base mesh elements

    """

    def __init__(self, node_x, node_y, cell_node):
        """Initializes Mesh object with node_x, node_y & cell_node"""

        # Store basic data as attributes
        self.node_x = node_x
        self.node_y = node_y
        self.cell_node = cell_node

        # Prepare additional mesh attributes
        self.__prepare_basic__()
        self.__prepare_edge_node__()
        self.__prepare_node_weights__()
        self.__prepare_tri_index__()

    def __prepare_basic__(self):
        """Command to prepare basic mesh geometry"""

        # Set null value
        null = -999

        # Count number of 2D cells and nodes
        self.nc2 = self.cell_node.shape[0]
        self.nv2 = self.node_x.shape[0]

        # Set node_x & node_y to 64 bit floating point arrays
        self.node_x = np.array(self.node_x).astype(np.float64)
        self.node_y = np.array(self.node_y).astype(np.float64)

        # Convert to integer Array for indexing purposes
        self.cell_node[np.isnan(self.cell_node)] = null
        self.cell_node = self.cell_node.astype(np.int32)

        # Identify triangles and repeat last node for triangles
        self.is_tri = self.cell_node[:, -1] == null
        self.is_quad = self.cell_node[:, -1] != null
        self.cell_node[self.is_tri, 3] = self.cell_node[self.is_tri, 2]

    def __prepare_edge_node__(self):
        """Command to define edge_nodes (nodes & cell associated with each half-edge)"""

        # Declare edge_node as tuple (n1, n2, idx)
        self.edge_node = \
            (
                self.cell_node[:, [0, 1, 2, 3]].flatten(),
                self.cell_node[:, [1, 2, 3, 0]].flatten(),
                np.repeat(np.arange(self.cell_node.shape[0]), 4),
                np.zeros((self.nc2 * 4,), dtype=bool)
            )

        # Delete 4th edge of triangle cells (repeated 3rd node)
        invalid = np.where(self.edge_node[0] == self.edge_node[1])[0]
        self.edge_node = tuple(np.delete(arr, invalid) for arr in self.edge_node)

    def __prepare_node_weights__(self):
        """Command to prepare node recovery weights for nodes local to each cell"""

        # Index nx, ny for each for 5 edges (4 angles)
        ii = [3, 0, 1, 2, 3, 0]
        nx = self.node_x[self.cell_node[:, ii]]
        ny = self.node_y[self.cell_node[:, ii]]

        # For triangle cells, set last corner to be _first corner
        nx[self.is_tri, 4:6] = nx[self.is_tri, 1:3]
        ny[self.is_tri, 4:6] = ny[self.is_tri, 1:3]

        # Calculate angles
        dx = nx[:, 1:] - nx[:, :-1]
        dy = ny[:, 1:] - ny[:, :-1]
        ds = np.sqrt(dx * dx + dy * dy)

        ang = np.arccos(-(dx[:, 1:] * dx[:, :-1] + dy[:, 1:] * dy[:, :-1]) / (ds[:, 1:] * ds[:, :-1]))

        # Set angle of repeated corner to 0
        ang[self.is_tri, 3] = 0

        # Sum angles for each node & divide by total angle
        ang_sum = np.bincount(self.cell_node.flatten(), weights=ang.flatten())
        self.weights = ang / ang_sum[self.cell_node]

    def __prepare_tri_index__(self):
        """Command to prepare cell and node indices of a triangular element mesh (splits quadrilateral elements)"""

        rep_mat = self.is_quad + 1

        tri_cell_node = self.cell_node.repeat(rep_mat, axis=0)
        tri_cell_index = np.arange(0, self.nc2).repeat(rep_mat)

        rep_index = np.where(tri_cell_index[:-1] == tri_cell_index[1:])[0]
        rep_index.shape = (rep_index.size, 1)

        tri_cell_node[rep_index + 1, [0, 1, 2]] = tri_cell_node[rep_index, [2, 3, 0]]

        self.tri_cell_node = tri_cell_node[:, 0:3]
        self.tri_cell_index = tri_cell_index

    def get_intersection_data(self, polyline):
        """
        Query to extract data for intersections(x) of a polyline and the mesh half-edges

        Parameters
        ----------
        polyline : 2D np.ndarray
            Polyline as [x, y] which intersects mesh half-edges.

        Returns
        -------
        x : tuple
            Intersection(x) data defined by coordinates & mesh cell index (x, y, ii)
        """

        # Basic preparation of data (avoids precision errors)
        polyline = np.array(polyline).astype(np.float64)

        # Prepare mesh segments as (x, y)
        c = (self.node_x[self.edge_node[0]], self.node_y[self.edge_node[0]])
        d = (self.node_x[self.edge_node[1]], self.node_y[self.edge_node[1]])

        # Prepare data to be returned  as tuple (x, y, cc)
        tup = (polyline[0, 0], polyline[0, 1], np.array([], dtype=np.int32))

        # Loop through each polyline segment
        ns = polyline.shape[0] - 1
        for ii in range(ns):
            # Get polyline segment intersections for segment ii
            a_ii, b_ii = tuple(polyline[ii]), tuple(polyline[ii + 1])
            t, u, lc, rc, lgi = get_intersection(a_ii, b_ii, c, d)

            # Index cell crossing parameters based on valid intersections
            t, u, lc, rc, cc = tuple(v[lgi] for v in (t, u, lc, rc, self.edge_node[2]))

            # Find index of null intersections (hits node)
            sort = np.lexsort((lc, t, cc,))
            node = sort[(u[sort] == 0) | (u[sort] == 1)]
            is_cell = cc[node][:-1] == cc[node][1:]
            is_space = t[node][:-1] != t[node][1:]
            keep = np.where(is_cell & is_space)[0]
            node = np.delete(node, np.hstack((keep, keep + 1)))

            # Remove null intersections
            t, lc, rc, cc = tuple(np.delete(v, node) for v in (t, lc, rc, cc))

            # Check number of valid intersections
            if t.size == 0:
                continue

            # Sort based on normalized distance along polyline AB
            sort = np.lexsort((cc, lc, t,))
            t, lc, rc, cc = tuple(v[sort] for v in (t, lc, rc, cc))

            # Further check for nodal intersections
            node = np.where((lc[:-1] & lc[1:]) | (rc[:-1] & rc[1:]))
            t, lc, rc, cc = tuple(np.delete(v, node) for v in (t, lc, rc, cc))

            # Find null cells
            null = np.where(((t[:-1] != t[1:]) & lc[1:]))[0] + 1
            null = np.insert(null, range(null.size), null)

            # Get cell indices
            cc = np.insert(cc, null, -999)
            if lc[0]:
                cc = np.insert(cc, 0, [-999, -999])
            else:
                cc = np.insert(cc, 0, cc[0])
            if rc[-1]:
                cc = np.insert(cc, cc.size, [-999, -999])
            else:
                cc = np.insert(cc, cc.size, cc[-1])
            cc = cc[range(0, cc.size, 2)]

            # Get unique edges between cells
            t = np.unique(t)
            x = a_ii[0] + t * (b_ii[0] - a_ii[0])
            y = a_ii[1] + t * (b_ii[1] - a_ii[1])

            # Add the  last points of the polyline
            x = np.insert(x, x.size, b_ii[0])
            y = np.insert(y, y.size, b_ii[1])

            # Concatenate data
            tup = tuple(np.hstack((tup[ii], arr)) for ii, arr in enumerate((x, y, cc)))

        return tup

    def get_grid_index(self, grid_x, grid_y):
        """
        Query to extract the mesh cell indices of each point in a grid

        Parameters
        ----------
        grid_x : 1D np.ndarray
            Horizontal grid point values
        grid_y : 1D np.ndarray
            Vertical grid point values

        Returns
        -------
        grid_index : 2D np.ndarray
            Mesh cell index for each point in grid
        """

        # Basic preparation of data (avoids precision errors)
        nx, ny = grid_x.size, grid_y.size
        grid_x = np.round(np.array(grid_x), 8).astype(np.float64)
        grid_y = np.round(np.array(grid_y), 8).astype(np.float64)

        # Prepare mesh segments as (x, y)
        c = (self.node_x[self.edge_node[0]], self.node_y[self.edge_node[0]])
        d = (self.node_x[self.edge_node[1]], self.node_y[self.edge_node[1]])

        # Segment dy for narrowing search
        dy = np.max(np.abs(d[1] - c[1]))

        # Declare mesh cell index array
        grid_index = np.zeros((ny, nx), dtype=np.int32) - 999

        # Loop through horizontal grid lines
        for ii in range(ny):
            a_ii, b_ii = (grid_x[0], grid_y[ii]), (grid_x[-1], grid_y[ii])

            # Get logical index of mesh segments in zone
            in_zone = (d[1] < (a_ii[1] + dy)) & (d[1] > (a_ii[1] - dy))

            # Filter mesh segments based on logical index
            c_ii = (c[0][in_zone], c[1][in_zone])
            d_ii = (d[0][in_zone], d[1][in_zone])

            # Get polyline segment intersections for horizontal ray
            t, u, lc, rc, lgi = get_intersection(a_ii, b_ii, c_ii, d_ii)

            # Index cell crossing parameters based on valid intersections
            t, u, lc, rc, cc = tuple(v[lgi] for v in (t, u, lc, rc, self.edge_node[2][in_zone]))

            # Find index of null intersections (hits node)
            sort = np.lexsort((lc, t, cc,))
            node = sort[(u[sort] == 0) | (u[sort] == 1)]
            is_cell = cc[node][:-1] == cc[node][1:]
            is_space = t[node][:-1] != t[node][1:]
            keep = np.where(is_cell & is_space)[0]
            node = np.delete(node, np.hstack((keep, keep + 1)))

            # Remove null intersections
            t, lc, rc, cc = tuple(np.delete(v, node) for v in (t, lc, rc, cc))

            # Check number of valid intersections
            if t.size == 0:
                continue

            # Sort based on normalized distance along polyline AB
            sort = np.lexsort((cc, lc, t,))
            t, lc, rc, cc = tuple(v[sort] for v in (t, lc, rc, cc))

            # Further check for nodal intersections
            node = np.where((lc[:-1] & lc[1:]) | (rc[:-1] & rc[1:]))
            t, lc, rc, cc = tuple(np.delete(v, node) for v in (t, lc, rc, cc))

            # Find null cells
            null = np.where(((t[:-1] != t[1:]) & lc[1:]))[0] + 1
            null = np.insert(null, range(null.size), null)

            # Get cell indices
            cc = np.insert(cc, null, -999)
            if lc[0]:
                cc = np.insert(cc, 0, [-999, -999])
            else:
                cc = np.insert(cc, 0, cc[0])
            if rc[-1]:
                cc = np.insert(cc, cc.size, [-999, -999])
            else:
                cc = np.insert(cc, cc.size, cc[-1])
            cc = cc[range(0, cc.size, 2)]

            # Get unique edges between cells
            xi = a_ii[0] + np.unique(t) * (b_ii[0] - a_ii[0])

            # Add the _first & last points of the polyline
            xi = np.insert(xi, [0, xi.size], [a_ii[0], b_ii[0]])

            # Find cell index based on xe(ii) < xm < xe(ii+1)
            xm = np.tile(grid_x, (cc.size, 1)).transpose()
            xe = np.tile(xi, (nx, 1))

            index = np.where((xe[:, :-1] < xm) & (xm < xe[:, 1:]))

            grid_index[ii, index[0]] = cc[index[1]]

        return grid_index

    def get_cell_index(self, xp, yp):
        """
        Query to extract the mesh cell indices of scatter points

        Parameters
        ----------
        xp : 1D np.ndarray
            X coordinate of scatter data
        yp : 1D np.ndarray
            Y coordinate of scatter data

        Returns
        -------
        index : 1D np.ndarray
            Mesh cell index for each scatter point
        """

        node_x = np.array(self.node_x).astype(np.float64)
        node_y = np.array(self.node_y).astype(np.float64)
        xp = np.array(xp, ndmin=1).astype(np.float64)
        yp = np.array(yp, ndmin=1).astype(np.float64)

        xp, yp = np.round(xp, 8), np.round(yp, 8)

        # three points and ids for triangles
        at = (node_x[self.cell_node[self.is_tri, 0]], node_y[self.cell_node[self.is_tri, 0]])
        bt = (node_x[self.cell_node[self.is_tri, 1]], node_y[self.cell_node[self.is_tri, 1]])
        ct = (node_x[self.cell_node[self.is_tri, 2]], node_y[self.cell_node[self.is_tri, 2]])
        it = np.arange(self.nc2)[self.is_tri]

        # four points and ids for quadrangles
        aq = (node_x[self.cell_node[self.is_quad, 0]], node_y[self.cell_node[self.is_quad, 0]])
        bq = (node_x[self.cell_node[self.is_quad, 1]], node_y[self.cell_node[self.is_quad, 1]])
        cq = (node_x[self.cell_node[self.is_quad, 2]], node_y[self.cell_node[self.is_quad, 2]])
        dq = (node_x[self.cell_node[self.is_quad, 3]], node_y[self.cell_node[self.is_quad, 3]])
        iq = np.arange(self.nc2)[self.is_quad]

        index = np.zeros(xp.size, dtype=np.int32) - 999
        for aa in range(xp.size):
            p = (xp[aa], yp[aa])

            # check ccw for triangles
            ccw = \
                (
                    is_ccw(p, at, bt),
                    is_ccw(p, bt, ct),
                    is_ccw(p, ct, at)
                )
            ii = np.where(np.all(ccw, axis=0))[0]

            if ii.size == 1:
                index[aa] = it[ii]

            # check ccw for quadrangles
            ccw = \
                (
                    is_ccw(p, aq, bq),
                    is_ccw(p, bq, cq),
                    is_ccw(p, cq, dq),
                    is_ccw(p, dq, aq)
                )
            ii = np.where(np.all(ccw, axis=0))[0]

            if ii.size == 1:
                index[aa] = iq[ii]

        return index

    def get_barycentric_weights(self, xp, yp, cell_index=None):
        pass



"""MATLAB style date and time functions"""

import datetime as dt
import numpy as np


def datetime(in_array, fmt='%d/%m/%Y %H:%M:%S'):
    """Converts date numbers or date strings to python datetime objects"""

    # convert input to np.array
    in_array = np.array(in_array, ndmin=1)

    # initialize conversion logic
    from_str = False
    from_flt = False
    from_int = False

    # determine conversion logic
    arr_type = in_array.dtype.type
    if arr_type is np.str_:
        from_str = True
    elif arr_type is np.float64:
        from_flt = True
    elif arr_type is np.int32:
        from_int = True

    # determine output array shape & size
    arr_shape = in_array.shape
    if from_int:
        if len(arr_shape) > 1:
            arr_shape = arr_shape[:-1]
        else:
            arr_shape = (1,)
    arr_size = int(np.prod(arr_shape))

    # flatten the input array
    if from_int:
        in_array = in_array.reshape((arr_size, in_array.shape[-1]))
    else:
        in_array = in_array.flatten()

    # initialize output array
    out_array = np.empty((arr_size,), dtype=np.object)

    # if from string -> datetime
    if from_str:
        for ii in range(arr_size):
            out_array[ii] = dt.datetime.strptime(in_array[ii], fmt)

    # if from float -> datetime
    if from_flt:
        epoch = dt.datetime(1970, 1, 1)
        in_array = in_array.astype(np.float64)
        for ii in range(arr_size):
            delta = dt.timedelta(seconds=in_array[ii])
            out_array[ii] = epoch + delta

    # if from int -> datetime
    if from_int:
        for ii in range(arr_size):
            out_array[ii] = dt.datetime(*list(in_array[ii]))

    # reshape the output array
    out_array = out_array.reshape(arr_shape)

    # handle output
    if out_array.size == 1:
        return out_array[0]
    else:
        return out_array


def datestr(in_array, fmt='%d/%m/%Y %H:%M:%S'):
    """Converts python datetime objects or date numbers to date strings"""

    # convert input to np.array
    in_array = np.array(in_array, ndmin=1)

    # convert everything to datetime
    arr_type = in_array.dtype.type
    if arr_type is not np.object_:
        in_array = datetime(in_array)
        in_array = np.array(in_array, ndmin=1)

    # get the array shape and size
    arr_shape = in_array.shape
    arr_size = np.prod(arr_shape)

    # flatten the input array
    in_array = in_array.flatten()

    # initialize output array
    str_check = in_array[0].strftime(fmt)
    d_type = '<U{:d}'.format(len(str_check))
    out_array = np.empty((arr_size,), dtype=d_type)

    # convert datetime to string
    for ii in range(in_array.size):
        out_array[ii] = in_array[ii].strftime(fmt)

    # reshape the output array
    out_array = out_array.reshape(arr_shape)

    # handle output
    if out_array.size == 1:
        return out_array[0]
    else:
        return out_array


def datenum(in_array, fmt='%d/%m/%Y %H:%M:%S'):
    """Converts python datetime objects or date strings to date numbers"""

    # convert input to np.array
    in_array = np.array(in_array, ndmin=1)

    # convert everything to datetime
    arr_type = in_array.dtype.type
    if arr_type is not np.object_:
        in_array = datetime(in_array, fmt=fmt)
        in_array = np.array(in_array, ndmin=1)

    # get the array shape and size
    arr_shape = in_array.shape
    arr_size = np.prod(arr_shape)

    # flatten the input array
    in_array = in_array.flatten()

    # initialize the output array
    out_array = np.empty((arr_size,), dtype=np.float64)

    # Convert datetime to number
    epoch = dt.datetime(1970, 1, 1)
    for ii in range(in_array.size):
        delta = in_array[ii] - epoch
        out_array[ii] = delta.total_seconds()

    # reshape the output array
    out_array = out_array.reshape(arr_shape)

    # handle output
    if out_array.size == 1:
        return out_array[0]
    else:
        return out_array


def datevec():
    """Python implementation of datevec"""
    pass


def convtime(in_array, to='py'):
    """Converts tuflow time to python time"""

    # Define epoch relative to python time
    fv_epoch = datenum((1990, 1, 1))

    # Convert input to datenum
    in_array = datenum(in_array)

    if to == 'py':
        return in_array*3600 + fv_epoch
    elif to == 'fv':
        return (in_array - fv_epoch)/3600


def timeit(function_call):

    name = function_call.__name__
    message = '{} took {}s to execute'

    def wrapper(*args, **kwargs):
        ts = dt.datetime.now().timestamp()
        output = function_call(*args, **kwargs)
        te = dt.datetime.now().timestamp()
        print(message.format(name, te - ts))

        return output

    return wrapper



"""A module defining all Extractor classes. Extractors are the primary objects for model result data extraction"""

from ast import Slice
import os
import re
from pathlib import Path
import numpy as np
import xarray as xr
import pandas as pd
from scipy import sparse
from inspect import getdoc
from abc import ABC, abstractmethod
from typing import Union
from types import GeneratorType
from datetime import datetime as dt
from dask.diagnostics import ProgressBar
from tqdm import tqdm
import warnings

# from netCDF4 import Dataset
from tfv.geometry import Mesh
from tfv.miscellaneous import *
from tfv.mldatetime import *

time_slice_err = [
    "`time_limits` optional argument must be a slice function",
    "Examples:",
    "   slice(0, 10) for the first 10 timesteps",
    "   slice('2020-01-01', '2020-02-01') to slice between dates",
    "For more help, please refer to Pandas `.loc` or Xarray `.isel` or `.sel` methods",
]


class Extractor(ABC):
    """
    A base class that defines the API for all model result subclasses. Examples of these subclasses
    might be model results such as a TUFLOW FV NetCDF file, a Hycom NetCDF file or a TUFLOW FV .dat file.
    """

    result_type = None

    def __init__(
        self,
        file: Union[Path, str, xr.Dataset],
        is_spherical: bool,
        lazy_load: bool,
        warmup: Union[str, pd.Timedelta],
    ):
        """Initializes Extractor object with a model results file i.e A TUFLOW FV netCDF4 results file."""

        # Store file path string as attribute
        self.file = file
        self.is_spherical = is_spherical

        # Convert warmup
        if isinstance(warmup, str):
            warmup = pd.Timedelta(warmup)

        # Prepare static Extractor attributes
        self.__prep_file_handle__(lazy_load, warmup)
        self.__prep_2d_geometry__()
        self.__prep_3d_geometry__()

    @abstractmethod
    def get_raw_data(self, variable: str, ii: int):
        """
        Query to extract raw data at a time step (if time-varying).

        Parameters
        ----------
        variable : string
            Name of time varying data set to be extracted.
        ii : integer
            The time vector index at which to extract the data.

        Returns
        -------
        data : np.ndarray
            The raw data as 1D or 2D numpy array
        """
        pass

    @abstractmethod
    def get_mask_vector(self, ii: int):
        """
        Query to extract an array that defines invalid model data.

        Parameters
        ----------
        ii : integer
            Time index at which to extract the stat array.

        Returns
        -------
        mask : np.ndarray
            Logical index, True if model cells/nodes are invalid (i.e dry cells).

        """
        pass

    @abstractmethod
    def get_z_layer_faces(self, ii: int):
        """
        Query to extract an array that defines the vertical layer faces of a 3D model.

        Parameters
        ----------
        ii : integer
            Time index at which to extract the vertical layer faces.

        Returns
        -------
        lfz : np.darray
            Vertical layer faces. If model is 2D returns None.

        """
        pass

    @abstractmethod
    def get_integral_data(self, ii: int, datum: str, limits: tuple):
        """
        Query to extract data for vertical integration at given time step. Principle data is the
        integral limit (z2 - z1) for each 2D model cell/node and dz for each 3D model cell/node.

        Parameters
        ----------
        ii : integer
            Time index at which to extract the vertical layer faces.
        datum : str
            {'sigma', 'depth', 'height', 'elevation'}
            Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
        limits : tuple
            Vertical depth-averaging limits (z1, z2) relative to vertical datum.

        Returns
        -------
        z_data : tuple, (z2_z1, dz)
            The elevation limits (z2 - z1) for each 2D cell/node & dz for each 3D cell/node

        """
        pass

    @abstractmethod
    def get_sheet_cell(
        self, variable: str, ii: int, datum="sigma", limits=(0, 1), z_data: tuple = None
    ):
        """
        Query to extract data in a 2D map format ('sheet') at model cell centroids for a given time step. If model
        data is 3D then it is depth-averaged according to the depth-averaging vertical datum and vertical limits.

        Parameters
        ----------
        variable : string
            Name of time varying data set to be extracted.
        ii : integer
            Time index at which to extract the data.
        datum : {'sigma', 'depth', 'height', 'elevation'}
            Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
        limits : tuple
            Vertical depth-averaging limits (z1, z2) relative to vertical datum.

        Other Parameters
        ----------------
        z_data : tuple, optional
            Vertical integration data returned by ```self.get_integral_data```

        Returns
        -------
        data : np.ndarray
            A 2D spatial 'sheet' of the relevant variable at time step ii.

        """
        pass

    @abstractmethod
    def get_sheet_node(
        self, variable: str, ii: int, datum="sigma", limits=(0, 1), z_data: tuple = None
    ):
        """
        Query to extract data in a 2D map format ('sheet') at model cell vertices for a given time step. If model
        data is 3D then it is depth-averaged according to the depth-averaging vertical datum and vertical limits.

        Parameters
        ----------
        variable : string
            Name of time varying data set to be extracted.
        ii : integer
            Time index at which to extract the data.
        datum : {'sigma', 'depth', 'height', 'elevation'}
            Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
        limits : tuple
            Vertical depth-averaging limits (z1, z2) relative to vertical datum.
        z_data : tuple, optional
            Vertical integration data returned by self.get_integral_data

        Returns
        -------
        data : np.ndarray
            A 2D spatial 'sheet' of the relevant variable at time step ii.

        """
        pass

    @abstractmethod
    def get_sheet_grid(
        self,
        variable: str,
        ii: int,
        grid_x: np.ndarray,
        grid_y: np.ndarray,
        datum="sigma",
        limits=(0, 1),
        z_data: tuple = None,
        grid_index: np.ndarray = None,
    ):
        """
        Query to extract data in a 2D map format ('sheet') at fixed grid points for a given time step. If model
        data is 3D then it is depth-averaged according to the depth-averaging vertical datum and vertical limits.

        Parameters
        ----------
        variable : string
            Name of time varying data set to be extracted.
        ii : integer
            Time index at which to extract the data.
        grid_x : 1D np.ndarray
            Horizontal grid point values
        grid_y : 1D np.ndarray
            Vertical grid point values
        datum : {'sigma', 'depth', 'height', 'elevation'}
            Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
        limits : tuple
            Vertical depth-averaging limits (z1, z2) relative to vertical datum.

        Other Parameters
        ----------------
        z_data : tuple, optional
            Vertical integration data returned by self.get_integral_data
        grid_index : 2D np.ndarray
            Mesh cell index for each grid point returned by self.get_grid_index

        Returns
        -------
        data  : 2D np.ndarray
            A gridded 2D spatial 'sheet' of the relevant variable at time step ii.

        """
        pass

    @abstractmethod
    def get_curtain_cell(
        self,
        variable: str,
        ii: int,
        polyline: np.ndarray,
        x_data: tuple = None,
        index: tuple = None,
    ):
        """
        Query to extract data in a 2D slice format ('curtain') at the cell centroids for a given time step. It does
        this along the polyline specified.

        Parameters
        ----------
        variable : string
            Name of time varying data set to be extracted.
        ii : integer
            Time index at which to extract the data.
        polyline : 2D np.ndarray
            Polyline as [x, y] used to slice 3D data.

        Other Parameters
        ----------------
        x_data : tuple
            Model edge intersection(x) data returned by self.get_intersections.
        index : tuple
            Curtain index data returned by self.get_curtain_cell_index.

        Returns
        -------
        data : np.ndarray
            A 2D slice 'curtain' of the relevant variable at time step ii.

        """
        pass

    @abstractmethod
    def get_curtain_edge(
        self,
        variable: str,
        ii: int,
        polyline: np.ndarray,
        x_data: tuple = None,
        index: tuple = None,
    ):
        """
        Query to extract data in a 2D slice format ('curtain') at the cell centroids for a given time step. It does
        this along the polyline specified.

        This function is currently not supported.
        """
        pass

    @abstractmethod
    def get_curtain_grid(
        self,
        variable: str,
        ii: int,
        polyline: np.ndarray,
        grid_x: np.ndarray,
        grid_y: np.ndarray,
        x_data: tuple = None,
        index: tuple = None,
        grid_index: np.ndarray = None,
    ):
        """
        Query to extract data in a 2D slice format ('curtain') at fixed grid points for a given time step. It does
        this along the polyline specified.

        Parameters
        ----------
        variable : string
            Name of time varying data set to be extracted.
        ii : integer
            Time index at which to extract the data.
        polyline : 2D np.ndarray
            Polyline as [x, y] used to slice 3D data.
        grid_x : 1D np.ndarray
            Horizontal grid point values
        grid_y : 1D np.ndarray
            Vertical grid point values

        Other Parameters
        ----------------
        x_data : tuple
            Model edge intersection(x) data returned by self.get_intersections.
        index : tuple
            Curtain index data returned by self.get_curtain_cell_index.
        grid_index : 2D np.ndarray
            Curtain cell index for each grid point

        Returns
        -------
        data : 2D np.ndarray
            A gridded 2D slice 'curtain' of the relevant variable at time step ii.

        """
        pass

    @abstractmethod
    def get_profile_cell(self, variable, ii, point, index=None):
        """
        Query to extract data as 1D vertical profile of cells at the given time step.
        It does this at the point specified.

        Parameters
        ----------
        variable : string
            Name of time varying data set to be extracted.
        ii : integer
            Time index at which to extract the data.
        point : tuple
            Point (x, y) of profile location.

        Other Parameters
        ----------------
        index : integer
            Index of cell which contains the point.

        Returns
        -------
        data : np.ndarray
            A 1D section of the vertical values of the relevant variable.

        """
        pass

    @abstractmethod
    def __prep_file_handle__(self):
        """Command which prepares the file handle for the extractor class"""

    @abstractmethod
    def __prep_time_vector__(self):
        """Command which prepares the result time stamp vector relative to python epoch"""

    @abstractmethod
    def __prep_2d_geometry__(self):
        """Command which prepares the result 2D mesh geometry"""

    @abstractmethod
    def __prep_3d_geometry__(self):
        """A command which prepares the result 3D mesh geometry"""


class FvExtractor(Extractor, Mesh):
    """
    Class that extracts data from a TUFLOW FV netCDF4 result file.

    Parameters
    ----------
    file : string
        Model result file path.

    Other Parameters
    ----------------
    is_spherical : bool
        True if model geometry defined in spherical coordinate reference system.

    Attributes
    ----------
    nc2 : int
        Number of 2D mesh cells
    nv2 : int
        Number of 2D mesh vertices
    is_tri : 1D np.ndarray
        Logical index of triangular elements
    is_quad : 1D np.ndarray
        Logical index of quadrilateral elements
    edge_node : tuple
        Tuple defining start node, end node and cell for each mesh half edge
    weights : 2D np.ndarray
        A (n, 4) array defining weighting of each cell gives to each mesh vertex
    tri_cell_node : 2D np.ndarray
        A (n, 3) array defining each mesh cell/element by three node indices
    tri_cell_index : 1D np.ndarray
        A (n,) array mapping triangular mesh elements to base mesh elements
    nc : netCDF4.Dataset
        Dataset object
    nz: np.ndarray
        A (nc2,) array defining number of vertical cells in each 2D model mesh cell
    idx2: np.ndarray
        A (nc3,) array defining the 2D model mesh cell index for each 3D cell
    idx3: np.ndarray
        A (nc2,) array defining the surface 3D model mesh cell index for each 2D model mesh cell
    idx4: np.ndarray
        A (nc3+nc2,) array defining the 2D model mesh cell index for each 3D vertical layer face
    wli: np.ndarray
        A (nc2,) array defining the surface vertical layer face index for each 2D model mesh cell
    bli: np.ndarray
        A (nc2,) array defining the bed vertical layer face index for each 2D model mesh cell
    """

    result_type = "Cell-centred TUFLOWFV output"

    def __init__(
        self,
        file: Union[Path, str, xr.Dataset],
        lazy_load: bool = True,
        is_spherical: bool = True,
        warmup: Union[str, pd.Timedelta] = "0D",
    ):
        super(FvExtractor, self).__init__(file, is_spherical, lazy_load, warmup)
        # self.__promote_coords__()

    @property
    def variables(self):
        return [
            x
            for x in self.ds.data_vars.keys()
            if "Time" in self.ds[x].dims
            if x not in ["ResTime", "stat", "layerface_Z"]
        ]

    # def _repr_(self):
    #     print(ds.__repr__())

    # def _repr_html_(self):
    #     display(self.ds)

    # def __getitem__(self, variable):
    #     return self.ds[variable]

    # def __setitem__(self, variable, array):
    #     self.ds[variable] = array

    def get_raw_data(self, variable: str, ii: int):
        # return self.ds.isel(Time=ii)[variable].values
        return self.ds[variable][ii, :].values

    def get_mask_vector(self, ii: int):
        return self.ds["stat"][ii, :].values == 0

    def get_z_layer_faces(self, ii: int):
        return self.ds["layerface_Z"][ii, :].values

    def get_integral_data(self, ii: int, datum: str, limits: tuple):
        lfz = self.get_z_layer_faces(ii)

        # Get water level (wl) and bed level (bl) of each 2D cell
        wl = lfz[self.wli]
        bl = lfz[self.bli]

        # Determine integral limits z1 and z2 for each 2D cell using wl, bl and the limits
        if datum == "sigma":
            z1 = limits[0] * (wl - bl) + bl
            z2 = limits[1] * (wl - bl) + bl
        elif datum == "height":
            z1 = limits[0] + bl
            z2 = limits[1] + bl
        elif datum == "depth":
            z1 = wl - limits[1]
            z2 = wl - limits[0]
        elif datum == "elevation":
            z1 = limits[0]
            z2 = limits[1]
        else:
            return None

        # Create integral limits, filtering z2 and z1 above and below water level or bed level
        z1 = np.minimum(np.maximum(z1, bl), wl)
        z2 = np.minimum(np.maximum(z2, bl), wl)

        # Squeeze out middle value of each vertical layer face
        lfz = np.maximum(lfz, z1[self.idx4])
        lfz = np.minimum(lfz, z2[self.idx4])

        # Get upper z layer face and lower z layer face for each 3D cell
        ul = np.delete(lfz, self.bli)
        ll = np.delete(lfz, self.wli)
        dz = ul - ll

        # Clean up integral limit (z2 - z1) to avoid division by zero
        z2_z1 = z2 - z1
        mask = z2_z1 == 0

        # Return integral limit of each 2D cell and dz of each 3D cell contained within integral limit
        return np.ma.masked_array(data=z2_z1, mask=mask, fill_value=np.nan), dz

    @Expression.decorator
    def get_sheet_cell(
        self,
        variable: str,
        ii: Union[int, str, pd.Timestamp],
        datum="sigma",
        limits=(0, 1),
        agg="mean",
        mask_dry: bool = True,
        z_data: tuple = None,
    ):
        # Convert timestep into integer
        ii = self._timestep_index(ii)

        # Get the raw data
        data = self.get_raw_data(variable, ii)
        if mask_dry:
            mask = self.get_mask_vector(ii)
        else:
            mask = np.zeros((self.nc2,), dtype=bool)

        # Check if data is 3D
        if data.size == self.nc3:
            # Get integral data for depth averaging
            if z_data is None:
                z_data = self.get_integral_data(ii, datum, limits)
            z2_z1, dz = z_data

            # Update stat vector with invalid limits
            mask = mask | z2_z1.mask

            if agg == "mean":
                # Integrate the data w.r.t z
                data = np.bincount(self.idx2, data * dz) * (1 / z2_z1)
            elif agg == "min":
                data = self._sparse_array_reduction(
                    data * (dz > 0).astype(int), np.minimum
                )
            elif agg == "max":
                data = self._sparse_array_reduction(
                    data * (dz > 0).astype(int), np.maximum
                )
            else:
                assert False, "agg should be equal to either `mean`, `max` or `min`"

        # Reshape stat vector for 2D sheets i.e bed mass
        if data.shape != mask.shape:
            n = data.shape[1]
            mask = np.tile(mask, (n, 1))
            mask = np.transpose(mask)

        # Return the data
        return np.ma.MaskedArray(data=data, mask=mask, fill_value=np.nan)

    @Expression.decorator
    def get_sheet_node(
        self,
        variable: str,
        ii: int,
        datum="sigma",
        limits=(0, 1),
        agg="mean",
        mask_dry: bool = True,
        z_data: tuple = None,
    ):

        # Convert timestep into integer
        ii = self._timestep_index(ii)

        # Get the raw data
        data = self.get_raw_data(variable, ii)
        if mask_dry:
            mask = self.get_mask_vector(ii)
        else:
            mask = np.zeros((self.nc2,), dtype=bool)

        # Check if data is 3D
        if data.size == self.nc3:
            # Get integral data for depth averaging
            if z_data is None:
                z_data = self.get_integral_data(ii, datum, limits)
            z2_z1, dz = z_data

            # Update stat vector for invalid limits (z2 - z1)
            mask = mask | z2_z1.mask

            if agg == "mean":
                # Integrate the data w.r.t z
                data = np.bincount(self.idx2, data * dz) * (1 / z2_z1)
            elif agg == "min":
                data = self._sparse_array_reduction(data, np.minimum)
            elif agg == "max":
                data = self._sparse_array_reduction(data, np.maximum)
            else:
                assert False, "agg should be equal to either `mean`, `max` or `min`"

        # Create copy of 2D node recovery weights
        weights = np.copy(self.weights)

        # Set weightings of invalid cells to zero
        weights[mask, :] = 0

        # Rescale weightings to account for discounted cells
        weights_sum = np.bincount(self.cell_node.flatten(), weights.flatten())
        mask = weights_sum == 0
        weights_sum[mask] = -999
        weights = weights / weights_sum[self.cell_node]

        if data.ndim == 1:
            # For each cell, calculate the weighted nodal data values
            tmp = np.tile(data, (4, 1)).transpose() * weights

            # Sum weighted data values for each node to get final vertex data
            data_node = np.bincount(self.cell_node.flatten(), tmp.flatten())
        else:
            data_node = np.empty((self.nv2, data.shape[1]), dtype=np.float64)
            for jj in range(data.shape[1]):
                # For each cell, calculate the weighted nodal data values
                tmp = np.tile(data[:, jj], (4, 1)).transpose() * weights

                # Sum weighted data values for each node to get final vertex data
                data_node[:, jj] = np.bincount(self.cell_node.flatten(), tmp.flatten())

            mask = np.tile(mask, (data.shape[1], 1))
            mask = np.transpose(mask)

        # Return the data
        return np.ma.MaskedArray(data=data_node, mask=mask, fill_value=np.nan)

    def get_sheet_grid(
        self,
        variable: str,
        ii: int,
        xg: np.ndarray,
        yg: np.ndarray,
        datum="sigma",
        limits=(0, 1),
        agg='mean',
        z_data: tuple = None,
        grid_index: np.ndarray = None,
    ):

        # Convert timestep into integer
        ii = self._timestep_index(ii)

        # Get grid index
        if grid_index is None:
            grid_index = self.get_grid_index(xg, yg)
        mask = np.equal(grid_index, -999)
        valid = np.equal(mask, False)

        # Index mesh data using grid index
        grid_data = np.ma.MaskedArray(
            data=np.zeros(grid_index.shape) * np.nan, fill_value=np.nan, mask=mask
        )
        grid_data[valid] = self.get_sheet_cell(variable, ii, datum, limits, agg, z_data)[
            grid_index[valid]
        ]

        # Return gridded data
        return grid_data

    @Expression.decorator
    def get_curtain_cell(
        self,
        variable: str,
        ii: int,
        polyline: np.ndarray,
        x_data: tuple = None,
        index: tuple = None,
    ):

        # Convert timestep into integer
        ii = self._timestep_index(ii)

        # Get prerequisite data
        mask = self.get_mask_vector(ii)
        data = self.get_raw_data(variable, ii)

        # Check if data is 3D
        assert data.size == self.nc3, "Data is not 3D"

        # Get edge intersection data
        if x_data is None:
            x_data = self.get_intersection_data(polyline)
        _, _, idx = x_data

        # Get curtain index data
        if index is None:
            index = self.get_curtain_cell_index(polyline, x_data)
        line_index, cell_index = index

        # Return curtain data
        return np.ma.MaskedArray(
            data=data[cell_index], mask=mask[idx[line_index]], fill_value=np.nan
        )

    @unsupported_decorator
    def get_curtain_edge(
        self, variable: str, ii: int, polyline: np.ndarray, x_data=None, index=None
    ):
        pass

    def get_curtain_grid(
        self,
        variable: str,
        ii: int,
        polyline: np.ndarray,
        xg,
        yg,
        x_data=None,
        index=None,
        grid_index=None,
    ):

        # Convert timestep into integer
        ii = self._timestep_index(ii)

        # Get grid index
        if grid_index is None:
            geo = self.get_curtain_cell_geo(ii, polyline, x_data)
            grid_index = Mesh(*geo).get_grid_index(xg, yg)
        mask = np.equal(grid_index, -999)
        valid = np.equal(mask, False)

        # Index mesh data using grid index
        grid_data = np.ma.MaskedArray(
            data=np.zeros(grid_index.shape) * np.nan, fill_value=np.nan, mask=mask
        )
        grid_data[valid] = self.get_curtain_cell(variable, ii, polyline, x_data, index)[
            grid_index[valid]
        ]

        # Return gridded data
        return grid_data

    @Expression.decorator
    def get_profile_cell(self, variable: str, ii: int, point: tuple, index=None):

        # Convert timestep into integer
        ii = self._timestep_index(ii)

        # Get the raw data
        data = self.get_raw_data(variable, ii)

        # Get 2D cell index
        if index is None:
            index = self.get_cell_index(point[0], point[1])
            if index < 0:
                print("Warning: point coordinate not inside model domain")

        # Allow profile cell to work for 2D data as well
        if "NumCells3D" in self.ds[variable].dims:
            # Index the data
            data = data[self.idx2 == index]

            # Repeat to get discrete elements
            insert = np.arange(0, data.size)
            data = np.insert(data, insert, data[insert])

        elif "NumCells2D" in self.ds[variable].dims:
            data = data[np.arange(0, self.nc2) == index]
        else:
            assert False, "Variable must include either NumCells3D or NumCells2D"
        return data

    def __prep_file_handle__(self, lazy_load: bool, warmup: pd.Timedelta):
        # Assert the file exists
        if isinstance(self.file, str):
            self.file = Path(self.file)
            assert Path(
                self.file
            ).exists(), f"No such file or directory: \n{self.file.as_posix()}"
        elif isinstance(self.file, Path):
            assert (
                self.file.exists()
            ), f"No such file or directory: \n{self.file.as_posix()}"

        # ToDO: Refactor this if statement
        single_file = any([isinstance(self.file, x) for x in [str, Path]])
        multi_file = any([isinstance(self.file, x) for x in [list, GeneratorType]])

        # Direct xarray object passthrough
        if isinstance(self.file, xr.Dataset):
            self.ds = self.file
            self.__prep_time_vector__()

        # Normal file open
        elif (lazy_load == False) & single_file:
            self.ds = xr.open_dataset(self.file, decode_cf=False)
            self.__prep_time_vector__()
            self.ds = _discard_warmup(self.ds, warmup)

        # Open as an out of memory dataset (single file)
        elif (lazy_load == True) & single_file:
            self.ds = xr.open_mfdataset([self.file], decode_cf=False)
            self.__prep_time_vector__()

        # Require individual file loop loading
        elif multi_file:
            self.ds = _open_mf_tfv_dataset(self.file, warmup=warmup)
            self.time_vector = pd.to_datetime(self.ds["Time"].values)
            self.nt = self.time_vector.size

        else:
            msg = [
                "Unclear file(s) type",
                "Please supply either a str/path, a list of files, or an xr.Dataset",
                "",
            ]
            assert False, "\n".join(msg)

    def __prep_time_vector__(self):
        # Define fv epoch relative to python epoch
        fv_epoch = pd.Timestamp(1990, 1, 1)

        # Prepare time vector relative to python epoch
        # This if statement is a future check for when xarray starts decoding FV results
        if isinstance(self.ds["ResTime"].values[0], np.datetime64):
            self.time_vector = pd.to_datetime(self.ds["ResTime"].values)
        else:
            self.time_vector = (
                pd.to_timedelta(self.ds["ResTime"].values, unit="h") + fv_epoch
            )
        self.nt = self.time_vector.size
        self.ds["Time"] = pd.to_datetime(self.time_vector)

    def __prep_2d_geometry__(self):
        # Get basic data from file
        node_x = self.ds["node_X"].values
        node_y = self.ds["node_Y"].values
        cell_node = self.ds["cell_node"].values - 1

        # Identify null node indices
        cell_node[cell_node == -1] = -999

        # Pass commonly used attributes, and drop unrequired geo from dataset
        self.cell_x = self.ds["cell_X"].values
        self.cell_y = self.ds["cell_Y"].values
        self.cell_z = self.ds["cell_Zb"].values
        self.cell_a = self.ds["cell_A"].values

        # Check if spherical
        if "spherical" in self.ds.attrs:
            if self.ds.attrs["spherical"] == "true":
                self.is_spherical = True
            else:
                self.is_spherical = False

        # Initialize 2d geometry as Mesh
        Mesh.__init__(self, node_x, node_y, cell_node)

    def __prep_3d_geometry__(self):
        # Get basic data from file
        self.nz = self.ds["NL"].values

        # Prepare variables to define 3D mesh
        index = np.arange(self.nz.size)
        self.idx2 = np.repeat(index, self.nz)
        self.idx3 = np.cumsum(self.nz) - self.nz
        self.idx4 = np.repeat(index, self.nz + 1)
        self.wli = self.idx3 + index
        self.bli = self.wli + self.nz
        self.nc3 = self.idx2.size

    def _timestep_index(self, ii):
        # Function to handle timestep argument
        # Convert incoming arg into a Pd.Timestamp
        if isinstance(ii, str):
            time = pd.Timestamp(ii)
        elif any([isinstance(ii, x) for x in [int, np.int64, np.int32]]):
            return ii  # Early return
        else:
            time = ii

        ii = np.argmin(np.abs(self.time_vector - time))

        return ii

    def _subset_dataset(self, time):
        """Helper function to subset datasets"""
        # Use sliced xarray dataset
        dsx = self.ds
        int_types = [int, np.int32, np.int64, np.int16]
        date_types = [pd.Timestamp, np.datetime64, str]
        try:
            if any([isinstance(time, x) for x in int_types]):
                dsx = dsx.isel(Time=[time])
            elif isinstance(time, slice):
                if isinstance(time.stop, int):
                    dsx = dsx.isel(Time=time)
                else:
                    dsx = dsx.sel(Time=time)
            elif any([isinstance(time, x) for x in date_types]):
                dsx = dsx.sel(Time=[time])
            elif isinstance(time, list):
                dsx = dsx.isel(Time=time)
            elif time is None:
                dsx
            else:
                dsx = dsx.sel(Time=time)
        except AttributeError:
            print(time_slice_err)

        return dsx

    def _sparse_array_reduction(self, data, function):
        aux = sparse.csr_matrix(
            (data, self.idx2, np.arange(data.size + 1)), (data.size, self.nc2)
        ).tocsc()

        cut = aux.indptr.searchsorted(data.size)
        reduced_array = np.empty(self.nc2)
        reduced_array.ravel()[:cut] = function.reduceat(aux.data, aux.indptr[:cut])

        return reduced_array

    def get_curtain_cell_index(self, polyline: np.ndarray, x_data: tuple = None):
        """
        Query to extract 3D cell indices of 2D vertical slice ('curtain') for a given polyline.

        Parameters
        ----------
        polyline : 2D np.ndarray
            Polyline as [x, y] used to slice 3D data.

        Other Parameters
        ----------------
        x_data : tuple
            Model edge intersection(x) data returned by self.get_intersections.

        Returns
        -------
        index : tuple
            Index data defined by line index & cell index (line_index, cell_index). The line index provides the 1D
            polyline segment indices for each cell in the 2D vertical slice. The cell index provides the 3D model cell
            indices for each cell in the 2D vertical slice.
        """

        # Get cell x_data
        if x_data is None:
            x_data = self.get_intersection_data(polyline)
        _, _, idx = x_data

        # Determine which cells in the polyline are valid
        is_valid = idx != -999

        # Prepare indexing variables for curtain
        idx = idx[is_valid]  # 2D cell index for each column in curtain
        nz = self.nz[idx]  # Number of 3D cells for each column in curtain
        n = int(np.sum(nz))  # Total number of 3D cells in curtain
        idx3 = np.cumsum(nz) - nz  # Index of top 3D cell for each column in curtain

        # Get 1D polyline index of the curtain cells (maps curtain cell to polyline segments)
        line_index = np.repeat(np.where(is_valid)[0], nz)

        # Get 3D cell index of the curtain cells (maps curtain cell to 3D model cell)
        cell_index = np.repeat(self.idx3[idx], nz) + (
            np.arange(n) - np.repeat(idx3, nz)
        )

        return line_index, cell_index

    def get_curtain_cell_geo(
        self,
        ii: int,
        polyline: np.ndarray,
        x_data: tuple = None,
        index: tuple = None,
        return_unit_vector=False,
    ):
        """
        Query to extract geometry data for a 2D vertical slice ('curtain') along a given polyline.

        Parameters
        ----------
        ii : integer
            Time index at which to extract the data.
        polyline : 2D np.ndarray
            Polyline as [x, y] used to slice 3D data.

        Other Parameters
        ----------------
        x_data : tuple
            Model edge intersection(x) data returned by self.get_intersections.
        index : tuple
            Curtain index data returned by self.get_curtain_cell_index.

        Returns
        -------
        geo : tuple
            A tuple containing the geometry of the 2D vertical slice as (node_x, node_y, cell_node).

        """

        # Get layer face z
        lfz = self.get_z_layer_faces(ii)

        # Get edge intersection(x) data
        if x_data is None:
            x_data = self.get_intersection_data(polyline)
        x, y, idx = x_data
        nic2 = len(idx)
        nl = self.nz
        coords = np.stack((x, y)).T

        # Get curtain index
        if index is None:
            index = self.get_curtain_cell_index(polyline, x_data)
        line_index, cell_index = index

        # Convert x & y into (m) if spherical
        if self.is_spherical:
            r_e = 6.371 * 10**6
            cf = np.pi / 180
            x = (r_e * np.cos(y * cf)) * (x * cf)
            y = r_e * (y * cf)

        # Prepare curtain mesh (x, y)
        dx = np.diff(x)
        dy = np.diff(y)
        ds = np.hypot(dx, dy)
        s = np.cumsum(ds)
        s = np.hstack((0, s))

        s1 = s[line_index]
        s2 = s[line_index + 1]
        z1 = np.delete(lfz, self.wli)[cell_index]
        z2 = np.delete(lfz, self.bli)[cell_index]

        # Get nodes and node indices that define each cell
        n = cell_index.size
        node_x = np.vstack((s1, s2, s2, s1)).ravel("F")
        node_y = np.vstack((z1, z1, z2, z2)).ravel("F")
        cell_node = np.arange(n * 4, dtype=np.int32).reshape(n, 4)

        if return_unit_vector == True:
            # Get unit vector
            norm = np.array((-np.diff(coords[:, 1]), np.diff(coords[:, 0]))).T
            unorm_tmp = np.zeros((norm.shape[0], 2))
            unorm_tmp[:, 0] = norm[:, 0] / np.hypot(norm[:, 0], norm[:, 1])
            unorm_tmp[:, 1] = norm[:, 1] / np.hypot(norm[:, 0], norm[:, 1])
            unorm = []
            for aa in range(nic2):
                i = idx[aa]
                ext = np.broadcast_to(unorm_tmp[aa, :], [nl[i], 2])
                unorm.extend([ext])
            unorm = np.vstack(unorm).T

            # Get tangent vector
            tang = np.array((np.diff(coords[:, 0]), np.diff(coords[:, 1]))).T
            utang_tmp = np.zeros((norm.shape[0], 2))
            utang_tmp[:, 0] = tang[:, 0] / np.hypot(tang[:, 0], tang[:, 1])
            utang_tmp[:, 1] = tang[:, 1] / np.hypot(tang[:, 0], tang[:, 1])
            utang = []
            for aa in range(nic2):
                i = idx[aa]
                ext = np.broadcast_to(utang_tmp[aa, :], [nl[i], 2])
                utang.extend([ext])
            utang = np.vstack(utang).T

            return node_x, node_y, cell_node, unorm, utang
        else:
            return node_x, node_y, cell_node

    def get_profile_cell_geo(self, ii: int, point: tuple, index=None):
        # Get layer face z
        lfz = self.get_z_layer_faces(ii)

        # Get 2D cell index
        if index is None:
            index = self.get_cell_index(point[0], point[1])

        # Index faces of cell
        lfz = lfz[self.idx4 == index]

        # Repeat to represent discrete elements
        insert = np.arange(1, lfz.size - 1)
        lfz = np.insert(lfz, insert, lfz[insert])

        return lfz

    # ToDO: Think of a way to better ingrain this feature in xarray, accounting for
    # variable length layerface_z, which makes it difficult
    def write_time_series_file(
        self, out_file: Union[Path, str], locations: dict, variables: list = None
    ):
        """
        Query that writes 2D & 3D point time series data from a TUFLOW FV netCDF4 results file.

        Parameters
        ----------
        out_file : string
            File path to write time series file.
        locations : dictionary
            Extraction points as dict(SITE_1=(x, y), SITE_2=(x, y)).
        variables : list
            List of variables to extract. Default is all variables
        """

        if isinstance(out_file, str):
            out_file = Path(out_file)

        # Check if out path exists
        # ToDO: Improve messaging
        if out_file.exists():
            user_input = input(out_file.name + " exists - enter Y/n to proceed:\n")
            while True:
                if user_input.upper() == "Y":
                    os.remove(out_file)
                    break
                elif user_input.upper() == "N":
                    print('Finished - no files have been modified')
                    return
                else:
                    user_input = input("Invalid user input please enter Y/N:\n")

        # Get cell indicies for each location
        idx = {}
        for name, point in locations.items():
            ind_2d = self.get_cell_index(point[0], point[1])[0]
            ind_3d = np.where(self.idx2 == ind_2d)[0]
            ind_lfz = np.where(self.idx4 == ind_2d)[0]
            idx[name] = (ind_2d, ind_3d, ind_lfz, point)
            if idx[name] == -999:
                print("WARNING: point '{}' is outside of model domain".format(name))

        dsg = self.ds[["ResTime"]]
        dsg = dsg.drop(["Time"])
        dsg.attrs[
            "Origin"
        ] = "Profile extracted from TUFLOWFV cell-centered output using `tfv` python tools"
        dsg.attrs["Type"] = "Profile cell from TUFLOWFV output"
        dsg.to_netcdf(out_file)

        for pt, (i2, i3, i4, ptc) in idx.items():
            dsx = self.ds.sel(NumCells2D=i2, NumCells3D=i3, NumLayerFaces3D=i4)
            dsx = dsx.drop(
                [
                    "Time",
                    "ResTime",
                    "cell_Nvert",
                    "NL",
                    "cell_node",
                    "cell_A",
                    "node_X",
                    "node_Y",
                    "node_Zb",
                    "idx2",
                    "idx3",
                    "node_NVC2",
                    "node_cell2d_idx",
                    "node_cell2d_weights",
                ],
                errors="ignore",
            )
            dsx = dsx.rename(
                dict(
                    cell_X="X",
                    cell_Y="Y",
                    cell_Zb="Z",
                    NumCells3D="NumLayers",
                    NumLayerFaces3D="NumLayerFaces",
                )
            )

            for var in dsx.data_vars.keys():
                if len(dsx[var].dims) == 1:
                    dsx[var] = dsx[var].expand_dims(dim="N1", axis=1)

            # Optional variable check
            if variables:
                dsx = dsx[["X", "Y", "Z", "stat", "layerface_Z"] + variables]

            # Add N1 dim to match TUFLOW-FV Output
            dsx["X"] = dsx["X"].expand_dims(dim="N1")
            dsx["Y"] = dsx["Y"].expand_dims(dim="N1")
            dsx["Z"] = dsx["Z"].expand_dims(dim="N1")

            dsx.attrs["Source file"] = (
                self.file if isinstance(self.file, str) else self.file.as_posix()
            )
            dsx.attrs["Created"] = dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            dsx.attrs["Loc. name"] = pt
            dsx.attrs["Loc. coords"] = ptc
            dsx.attrs[
                "Origin"
            ] = "Profile extracted from TUFLOWFV cell-centered output using `tfv` python tools"
            dsx.attrs["Type"] = "Profile cell from TUFLOWFV output"

            wj = dsx.to_netcdf(out_file, mode="a", group=pt, compute=False)
            with ProgressBar(minimum=5):
                print(f"Writing location: {pt}")
                wj.compute()
        print("Finished")

    def write_data_to_ascii(
        self,
        out_file: Union[Path, str],
        data: np.ndarray,
        resolution: float,
        precision: int = 2,
        bbox: list = None,
        grid_index: np.ndarray = None,
    ):
        """
        Query data at cell centroids to gridded raster ASCII file.

        Parameters
        ----------
        out_file : string
            File path of .asc file to write data to.
        data : 1D np.ndarray
            Cell centred TUFLOW FV data as numpy array.
        resolution : float
            Grid resolution at which to output data.
        precision : int
            Output precision as number of decimal places.
        bbox : list, optional
            The bounding box to trim the data with as [left, bottom, right, top].
        grid_index : 2D np.ndarray, optional
            Mesh cell index for each grid point returned by self.get_grid_index

        """

        # Calculate grid parameters
        if bbox is None:
            x_min, x_max = self.node_x.min(), self.node_x.max()
            y_min, y_max = self.node_y.min(), self.node_y.max()
        else:
            x_min, x_max = bbox[0], bbox[2]
            y_min, y_max = bbox[1], bbox[3]

        dx = x_max - x_min
        dy = y_max - y_min

        nc = int(np.ceil(dx / resolution))
        nr = int(np.ceil(dy / resolution))

        xg = np.linspace(x_min, x_max, nc)
        yg = np.linspace(y_min, y_max, nr)

        # Get grid index & flip (x increasing, y decreasing)
        if grid_index is None:
            grid_index = self.get_grid_index(xg, yg)
        grid_index = np.flipud(grid_index)
        valid = grid_index != -999

        # Index mesh data using grid index
        grid_data = np.ones(grid_index.shape) * np.nan
        grid_data[valid] = data[grid_index[valid]]
        grid_data[np.isnan(grid_data)] = -999

        # Specify header
        header = [
            "ncols {:d}".format(nc),
            "nrows {:d}".format(nr),
            "xllcorner {:.7f}".format(x_min - resolution / 2),
            "yllcorner {:.7f}".format(y_min - resolution / 2),
            "cellsize {:.7f}".format(resolution),
            "NODATA_value {:d}".format(-999),
        ]

        # Write ASCII file
        with open(out_file, "w") as f:
            f.write("\n".join(header) + "\n")
            np.savetxt(f, grid_data, "%.{}f".format(precision), delimiter=" ")

    # Inherit doc strings (needs to be done a better way with decorator as per matplotlib)
    get_raw_data.__doc__ = Extractor.get_raw_data.__doc__
    get_mask_vector.__doc__ = Extractor.get_mask_vector.__doc__
    get_z_layer_faces.__doc__ = Extractor.get_z_layer_faces.__doc__
    get_integral_data.__doc__ = Extractor.get_integral_data.__doc__

    get_sheet_cell.__doc__ = getdoc(Extractor.get_sheet_cell)
    get_sheet_node.__doc__ = Extractor.get_sheet_node.__doc__
    get_sheet_grid.__doc__ = Extractor.get_sheet_grid.__doc__

    get_curtain_cell.__doc__ = Extractor.get_curtain_cell.__doc__
    get_curtain_edge.__doc__ = Extractor.get_curtain_edge.__doc__
    get_curtain_grid.__doc__ = Extractor.get_curtain_grid.__doc__
    get_profile_cell.__doc__ = Extractor.get_profile_cell.__doc__


def _open_mf_tfv_dataset(
    files: Union[list, GeneratorType], warmup: pd.Timedelta = None
):
    """ """
    fv_epoch = pd.Timestamp(1990, 1, 1)

    # Always sort file list
    if isinstance(files, GeneratorType):
        files = list(files)
    files.sort()

    # Loop through each file, chop warmup
    ds_set = []
    for f in files:
        ds = xr.open_mfdataset([f], decode_cf=False, chunks={"Time": 10})

        tvec = pd.to_timedelta(ds["ResTime"].values, unit="h") + fv_epoch
        ds["Time"] = tvec

        # Take out warmup period
        if warmup:
            ds = _discard_warmup(ds, warmup)

        ds_set.append(ds)

    ds = xr.concat(ds_set, dim="Time", data_vars="minimal")

    # Drop duplicates
    _, index = np.unique(ds["Time"], return_index=True)

    if len(index) < len(ds["Time"].values):
        print("Warning: Dropping duplicate times")
    ds = ds.isel(Time=index)
    ds = ds.sortby("Time")

    return ds


def _discard_warmup(ds: xr.Dataset, warmup: pd.Timedelta):
    ts = ds["Time"][0] + warmup
    return ds.sel(Time=slice(ts, None))


def _strip_dataset(ds):
    dst = xr.Dataset()

    keep_vars = ("ResTime", "layerface_Z", "stat")
    for dv in ds.data_vars.keys():
        if "Time" in ds[dv].dims:
            if dv in keep_vars:
                dst = dst.assign({dv: ds[dv]})
        else:
            dst = dst.assign({dv: ds[dv]})

    dst["stat"][:] = -1
    return dst



import setuptools

with open("README.md", "r") as rm:
    long_description = rm.read()

setup = \
    {
        'name': "tfv",
        'version': '1.0.1',
        'description': "Post processing tools for TUFLOW FV results",
        'long_description': long_description,
        'long_description_content_type': "text/markdown",
        'url': "https://gitlab.com/TUFLOW/tfv",

        'author': "TUFLOW Support",
        'author_email': "support@tuflow.com",

        'packages': ['tfv'],

        'classifiers':
            [
                "Programming Language :: Python :: 3",
                "License :: OSI Approved :: MIT License",
                "Operating System :: OS Independent",
            ],

        'python_requires': '>=3.9',
        'install_requires':
            [
                'numpy>=1.19.0',
                'matplotlib>3.2.2',
                'netCDF4>=1.5.3',
                'xarray>=v2022.03.0',
                'dask>=2022.01.0',
                'scipy>=1.6.0',
                'tqdm>=4.50.0'
            ],
        'extras_require':{
            'viewer': ["PyQt5>=5.15.0"]
        }
    }

setuptools.setup(**setup)



"""A module defining all objects used to visualise model result data extracts"""
import os
import time
import numpy as np
from tfv.mldatetime import *

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.animation import FFMpegWriter
from matplotlib.colorbar import Colorbar
from matplotlib.text import Text

# Check to see if PyQt5 is available
import_err_msg = "\n".join([
    "You must additionally install `PyQt5` to use the Viewer module",
    "Please install this using your relevant python package distribution",
    "PyPi - `pip install pyqt5` - https://pypi.org/project/PyQt5/ ",
    "Anaconda - `conda install -c anaconda pyqt` - https://anaconda.org/anaconda/pyqt",
    "Conda-forge - conda install -c conda-forge pyqt` - https://anaconda.org/conda-forge/pyqt"
])

import pkg_resources
try:
    pkg_resources.get_distribution('PyQt5')
except pkg_resources.DistributionNotFound:
    raise ImportError(import_err_msg)
else:
    HAS_PYQT5 = True

if HAS_PYQT5:
    from PyQt5.QtWidgets import QSlider, QWidget, QHBoxLayout, \
                                QVBoxLayout, QLabel, QPushButton
    from PyQt5.QtGui import QIcon
    from PyQt5.QtCore import Qt

# configure MPL
mpl.use('Qt5Agg')
plt.interactive(True)


class Slider:
    """
    A class for a single slider object placed in the MPL toolbar, built with PyQt5.

    Parameters
    ----------
    figure : tuple
        Size of underlying figure as (width, height) in specified units

    Other Parameters
    ----------------
    range : np.ndarray
        range of values that slider selects from
    step : integer
        The step size of the slider, i.e. how manly values it skips
    wait : float
        The wait time in seconds between each step when in 'play' mode
    format : {'num', 'hms', '%d/%m/%Y %H:%M:%S', '%d-%m-%Y', '%H:%M:%S %d-%m-%Y', ...}
        The slider value display format. Any custom datetime format string can be passed in addition
        to 'num' for the python date time as a number and 'hms' for total hours, minutes and seconds.

    Attributes
    ----------
    value : float
        The current value of the slider
    index : integer
        The current index of the slider
    text : string
        The display text of the slider
    """

    # Constructor functions
    def __init__(self,
                 figure,
                 range=np.arange(100),
                 step=1,
                 wait=0.01,
                 format='num'):
        # protected attributes
        self._figure = figure
        self._range = range

        self._value = None
        self._index = None
        self._text = ''
        self._playing = False
        self._callbacks = []

        # public attributes
        self.step = step
        self.wait = wait
        self.format = format

        # build the user interface
        self.__build_ui__()

        # set defaults
        self.value = 0
        self.index = 0

    def __build_ui__(self):
        """Method to build the User Interface (UI)"""

        # get the navigation widget handle
        nav_widget = self._figure.canvas.parent().children()[1]

        # add a separator to navigation widget
        nav_widget.addSeparator()

        # create and set layout of display widget
        self.display_widget = QWidget()
        self.display_layout = QVBoxLayout()
        self.display_widget.setLayout(self.display_layout)

        # format the display widget and layout
        self.display_widget.setContentsMargins(2, 2, 2, 2)
        self.display_layout.setContentsMargins(2, 2, 2, 2)
        self.display_widget.setMinimumWidth(125)
        self.display_widget.setMaximumWidth(125)

        # add the display widget to the navigation widget
        nav_widget.addWidget(self.display_widget)

        # create and set layout of label widget
        self.label_widget = QLabel(self._text)
        self.label_widget.setAlignment(Qt.AlignCenter)

        # add the label widget to the display layout
        self.display_layout.addWidget(self.label_widget)

        # create the slider widget
        self.slider_widget = QSlider(Qt.Horizontal)

        # format the slider widget
        self.slider_widget.setMinimum(0)
        self.slider_widget.setMaximum(len(self.range) - 1)

        # add the slider widget to the display layout
        self.display_layout.addWidget(self.slider_widget)

        # create and set layout of the buttons widget
        self.buttons_widget = QWidget()
        self.buttons_layout = QHBoxLayout()
        self.buttons_widget.setLayout(self.buttons_layout)

        # format the buttons widget and layout
        self.buttons_widget.setContentsMargins(2, 2, 2, 2)
        self.buttons_layout.setContentsMargins(2, 2, 2, 2)
        self.buttons_widget.setMinimumWidth(125)
        self.buttons_widget.setMaximumWidth(125)

        # add the buttons widget to the navigation widget
        nav_widget.addWidget(self.buttons_widget)

        # add the _first button
        self.first_button = QPushButton('|<')
        self.first_button.setMinimumWidth(20)
        self.first_button.setMaximumWidth(20)
        self.first_button.setToolTip('First Timestep')
        self.buttons_layout.addWidget(self.first_button)

        # add the previous button
        self.prev_button = QPushButton('<')
        self.prev_button.setMinimumWidth(20)
        self.prev_button.setMaximumWidth(20)
        self.prev_button.setToolTip('Previous Timestep')
        self.buttons_layout.addWidget(self.prev_button)

        # add the _next button
        self.next_button = QPushButton('>')
        self.next_button.setMinimumWidth(20)
        self.next_button.setMaximumWidth(20)
        self.next_button.setToolTip('Next Timestep')
        self.buttons_layout.addWidget(self.next_button)

        # add the _last button
        self.last_button = QPushButton('>|')
        self.last_button.setMinimumWidth(20)
        self.last_button.setMaximumWidth(20)
        self.last_button.setToolTip('Last Timestep')
        self.buttons_layout.addWidget(self.last_button)

        # add the play button
        self.play_button = QPushButton()
        self.play_button.setMinimumWidth(20)
        self.play_button.setMaximumWidth(20)
        self.play_button.setToolTip('Play Through Timesteps')
        self.buttons_layout.addWidget(self.play_button)

        # set the play button icon
        image_path = os.path.dirname(__file__) + "\\icons\\play_button.png"
        self.play_button.setIcon(QIcon(image_path))

        # connect slots to signals
        self.slider_widget.valueChanged.connect(self._changed)
        self.first_button.pressed.connect(self._first)
        self.prev_button.pressed.connect(self._prev)
        self.next_button.pressed.connect(self._next)
        self.last_button.pressed.connect(self._last)
        self.play_button.pressed.connect(self._toggle)

    # Protected member functions
    def _synchronise(self):
        # set the value of the slider widget
        self.slider_widget.setValue(self.index)

        # update the text display above the slider
        if self.format == 'int':
            self._text = str(self.index)
        elif type(self.value) == np.int32:
            self._text = str(self.index)
        else:
            self._text = self.value.strftime(self.format)
        self.label_widget.setText(self._text)

        # pass new value to call backs
        for callback in self._callbacks:
            callback(self.value)

    def _changed(self):
        self.index = self.slider_widget.value()

    def _first(self):
        self.index = 0

    def _prev(self):
        self.index = self.index - self.step

    def _next(self):
        self.index = self.index + self.step

    def _last(self):
        self.index = len(self.range) - 1

    def _toggle(self):
        path = os.path.dirname(__file__)
        if not self._playing:
            # _toggle the icon
            image_path = path + "\\icons\\stop_button.png"
            self.play_button.setIcon(QIcon(image_path))

            # _toggle flag
            self._playing = True

            # switch off interaction
            plt.interactive(False)

            # enter update loop
            while self._playing:
                self._next()
                self._redraw()
                time.sleep(self.wait)
        else:
            # _toggle the icon
            image_path = path + "\\icons\\play_button.png"
            self.play_button.setIcon(QIcon(image_path))

            # _toggle flag
            self._playing = False

            # switch on interaction
            plt.interactive(True)

    def _redraw(self):
        for axes in self._figure.axes:
            # Redraw the only items in frame
            axes.redraw_in_frame()

            # Draw x grid-lines
            x_grid = axes.get_xgridlines()
            for line in x_grid:
                axes.draw_artist(line)

            # Draw y grid-lines
            y_grid = axes.get_ygridlines()
            for line in y_grid:
                axes.draw_artist(line)

        # self.figure.canvas.update()
        self._figure.canvas.update()
        self._figure.canvas.flush_events()

    # Properties
    @property
    def figure(self):
        return self._figure

    @property
    def range(self):
        return self._range

    @range.setter
    def range(self, range):
        # set the protected range
        self._range = range

        # update slider limits
        min_val, max_val = 0, len(range) - 1
        self.slider_widget.setMinimum(min_val)
        self.slider_widget.setMaximum(max_val)

        # reset the value
        self.value = self._value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        # set the index based on nearest value in range
        if value == 0:
            value = self.range[0]
        self.index = np.argmin(np.abs(self.range - value))

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, index):
        # check upper and lower bounds
        min_val, max_val = 0, len(self.range) - 1
        if index > max_val:
            index = max_val
        if index < min_val:
            index = min_val

        if self.index != index or self.value != self.range[index]:
            self._index = index
            self._value = self.range[index]
            self._synchronise()

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, text):
        self._text = text
        self.label_widget.setText(text)

    # Public member functions
    def connect(self, function):
        """Connect function or handler value _changed event"""
        self._callbacks.append(function)

    def disconnect(self, function):
        """Disconnect function or handler value _changed event"""
        self._callbacks.remove(function)


class Viewer:
    """
    Class to explore/view time varying model result data

    Other Parameters
    ----------------
    size : tuple
        Size of underlying figure as (width, height) in specified units
    units : {'mm', 'cm' or 'm'}
        Figure units
    step : integer
        The step size of the slider, i.e. how manly values it skips
    wait : float
        The wait time in seconds between each step when in 'play' mode
    format : {'num', 'hms', '%d/%m/%Y %H:%M:%S', '%d-%m-%Y', '%H:%M:%S %d-%m-%Y', ...}
        The slider value display format. Any custom datetime format string can be passed in addition
        to 'num' for the python date time as a number and 'hms' for total hours, minutes and seconds.
    display : boolean
        Switch to display current time as text using the specified format. Default True.

    Attributes
    ----------
    figure : matplotlib.pyplot.Figure
        Underlying matplotlib figure object
    time_vector : np.ndarray
        Time vector associated with display
    time_current : np.float64
        Current time of display
    text : string
        Text of the display
    """

    __register__ = list()

    # Constructor function
    def __init__(self,
                 size=(240, 120),
                 units='mm',
                 step=1,
                 wait=0.01,
                 format='%d/%m/%Y %H:%M:%S',
                 display=True):
        """Initializes Viewer object with figure size, figure units & title time format"""

        # add self to the viewer register
        self.__register__.append(self)

        # public attributes
        self.figure = None
        self.size = size
        self.units = units
        self.axes = []

        # protected attributes
        self._slider_bar = None
        self._title_axes = None
        self._title_obj = None

        # set up the MPL figure
        sf_map = {'mm': 12 / 300, 'cm': 12 / 30, 'm': 12 / 0.3}

        sf = sf_map[self.units]
        w = self.size[0] * sf
        h = self.size[1] * sf

        self.figure = plt.figure(figsize=(w, h))

        # add the slider
        self._slider_bar = Slider(self.figure,
                                  step=step,
                                  wait=wait,
                                  format=format)

        # create title axes and add text for title
        rec = [0.30, 0.96, 0.40, 0.04]
        self._title_axes = self.figure.add_axes(rec)
        self._title_axes.xaxis.set_visible(False)
        self._title_axes.yaxis.set_visible(False)
        self._title_axes.set_navigate(False)

        for spine in self._title_axes.spines.values():
            spine.set_color('white')

        self._title_axes.set_xlim(0, 2)
        self._title_axes.set_ylim(0, 2)

        self._title_obj = Text(1,
                               1,
                               '',
                               fontsize=12,
                               horizontalalignment='center',
                               verticalalignment='center')
        self._title_axes._add_text(self._title_obj)

        # connect display text to the slider bar
        self._slider_bar.connect(self._set_text)

        # set text and show_time
        self.text = self._slider_bar.text
        self.show_time = display

    # Properties
    @property
    def time_current(self):
        return self._slider_bar.value

    @time_current.setter
    def time_current(self, time_current):
        self._slider_bar.value = time_current

    @property
    def time_vector(self):
        return self._slider_bar.range

    @time_vector.setter
    def time_vector(self, time_vector):
        # if np.all(self.time_vector == np.arange(100)):
        self._slider_bar.range = time_vector
        # else:
        #     self._slider_bar.range = np.arange(t1, t2 + dt, dt)

    @property
    def step(self):
        return self._slider_bar.step

    @step.setter
    def step(self, step):
        self._slider_bar.step = step

    @property
    def wait(self):
        return self._slider_bar.wait

    @wait.setter
    def wait(self, wait):
        self._slider_bar.wait = wait

    @property
    def time_fmt(self):
        return self._slider_bar.format

    @time_fmt.setter
    def time_fmt(self, time_fmt):
        self._slider_bar.format = time_fmt

    @property
    def text(self):
        return self._slider_bar.text

    @text.setter
    def text(self, text):
        self._title_obj.set_text(text)
        self._slider_bar.text = text

    @property
    def show_time(self):
        return self._title_axes.get_visible()

    @show_time.setter
    def show_time(self, show_time):
        self._title_axes.axes.set_visible(show_time)

    # protected member functions
    def _set_text(self, dummy):
        self.text = self._slider_bar.text

    # Public member functions
    def connect(self, function):
        """Subscribes callback to slider"""
        self._slider_bar.connect(function)

    def disconnect(self, function):
        """Removes callback subscription"""
        self._slider_bar.disconnect(function)

    def print(self, file_path, format='png', dpi=300):
        """
        Prints Viewer objects figure to image file.

        Parameters
        ---------
        file_path : string
            File path to output image file
        format : string
            File format i.e 'png', 'pdf', 'jpg'
        dpi : integer
            Resolution in dots per inch (dpi)
        """

        self.figure.savefig(file_path, format=format, dpi=dpi)

    def animate(self, file_path, ts, te, dt, fps=10, dpi=150):
        """
        Writes animation file for specified time interval.

        Parameters
        ----------
        file_path : string
            File path to output animation
        ts : float
            Start time as python time stamp
        te : float
            End time as python time stamp
        dt : float
            Time increment at which to save frame

        Other Parameters
        ----------------
        fps : integer
            Frames per second
        dpi : integer
            Resolution in dots per inch (dpi)
        """

        # Loop through time
        nt = int((te - ts) / dt) + 1
        writer = FFMpegWriter(fps=fps)
        with writer.saving(self.figure, file_path, dpi):
            for ii in range(nt):
                self.time_current = (ts + ii * dt)
                writer.grab_frame()


class ColourBar(Colorbar):
    """
    A wrapper around the matplotlib colorbar, simplifying for neater plots.

    Parameters
    ----------
    patch : matplotlib.collections.PatchCollection object
        matplotlib object from which to take colour limits and normalization

    Other Parameters
    ----------------
    location : {'left', 'right', 'bottom', 'top'}
        Size of underlying figure as (width, height) in specified units
    offset : float
        Offset from axes as fraction of total figure width or height
    thickness : float
        Thickness of bar as fraction of total figure width or height
    label : string
        String to label the colour bar.
    """

    def __init__(self,
                 patch,
                 location='bottom',
                 offset=0.075,
                 thickness=0.010,
                 label=''):

        # Get target axes handle
        if type(patch).__name__ == 'TriContourSet':
            target = patch.ax
        else:
            target = patch.axes

        if offset > 0:
            offset_1 = offset
            offset_2 = 0
        else:
            offset_1 = 0
            offset_2 = offset

        # Determine rectangles for target axes & colour bar axes
        rec = target.get_position().extents
        if location == 'bottom':
            rec1 = [
                rec[0], rec[1] + offset_1, rec[2] - rec[0],
                rec[3] - rec[1] - offset_1
            ]
            rec2 = [rec[0], rec[1] + offset_2, rec[2] - rec[0], thickness]
        elif location == 'top':
            rec1 = [
                rec[0], rec[1], rec[2] - rec[0], rec[3] - rec[1] - offset_1
            ]
            rec2 = [rec[0], rec[3] - offset_2, rec[2] - rec[0], thickness]
        elif location == 'left':
            rec1 = [
                rec[0] + offset_1, rec[1], rec[2] - rec[0] - offset_1,
                rec[3] - rec[1]
            ]
            rec2 = [rec[0] + offset_2, rec[1], thickness, rec[3] - rec[1]]
        elif location == 'right':
            rec1 = [
                rec[0], rec[1], rec[2] - rec[0] - offset_1, rec[3] - rec[1]
            ]
            rec2 = [rec[2] - offset_2, rec[1], thickness, rec[3] - rec[1]]

        target.set_position(rec1)
        axes = target.figure.add_axes(rec2)

        # Set orientation
        if location == 'bottom' or location == 'top':
            orientation = 'horizontal'
        elif location == 'left' or location == 'right':
            orientation = 'vertical'

        # Initialize Colorbar
        super(ColourBar, self).__init__(axes,
                                        mappable=patch,
                                        orientation=orientation)

        # Finish formatting
        if orientation == 'horizontal':
            axes.xaxis.set_ticks_position(location)
            axes.xaxis.set_label_position(location)
            axes.set_xlabel(label)
        elif orientation == 'vertical':
            axes.yaxis.set_ticks_position(location)
            axes.yaxis.set_label_position(location)
            axes.set_ylabel(label)
